@@ -58,7 +58,6 @@
 
 #include "mold.h"
 #include <tbb/parallel_for_each.h>
-#include <tbb/parallel_sort.h>
 
 namespace mold {
 
@@ -620,8 +619,8 @@ static std::vector<Compunit> read_compunits(Context<E> &ctx) {
 
   // Uniquify elements because GCC 11 seems to emit one record for each
   // comdat group which results in having a lot of duplicate records.
-  tbb::parallel_for_each(cus, [&](Compunit &cu) {
-    sort(cu.nametypes);
+  tbb::parallel_for_each(cus, [](Compunit &cu) {
+    ranges::stable_sort(cu.nametypes);
     remove_duplicates(cu.nametypes);
   });
 
