@@ -12,4 +12,4 @@ echo 'VER1 { foo; };' > $t/b.ver
 $CC -B. -shared -o $t/c.so $t/a.o -Wl,--version-script=$t/b.ver -flto
 readelf --symbols $t/c.so > $t/log
 
-grep -Fq 'foo@@VER1' $t/log
+grep -F 'foo@@VER1' $t/log
